#region License
/* ****************************************************************************
 * Copyright (c) Llewellyn Pritchard. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. 
 * A copy of the license can be found in the License.html file at the root of this distribution. 
 * By using this source code in any fashion, you are agreeing to be bound by the terms of the 
 * Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 * ***************************************************************************/
#endregion

using System;
using System.Collections.Generic;
using System.Text;

using Mast = Microsoft.Scripting.Ast.Ast;
using Microsoft.Scripting.Ast;
using Microsoft.Scripting;
using System.Reflection;
using IronLisp.Compilers;
using System.Collections;
using Microsoft.Scripting.Generation;
using System.IO;
using System.Diagnostics;
using Microsoft.Scripting.Utils;
using IronLisp.Runtime;
using Microsoft.Scripting.Actions;

using MSA = Microsoft.Scripting.Ast;


namespace IronLisp.Ast
{
  public abstract class SpecialExpression : Expression
  {
    public sealed override MSA.Expression GetCons(CodeBlock cb)
    {
      Debug.Fail("OBSOLETE!!!!");
      throw new Exception("Cant you read above?");
    }

    protected static MSA.Expression MakeClosure(CodeBlock cb)
    {
      return Mast.Call(null,
        typeof(Closure).GetMethod("Make"), Mast.CodeContext(), Mast.CodeBlockExpression(cb, false, false), Mast.Constant(cb.Name));
    }

    protected static string GetFullname(CodeBlock cb, string name)
    {
      if (cb == null || cb.IsGlobal)
      {
        return name;
      }
      else
      {
        return cb.Name + "::" + name;
      }
    }

    protected static MSA.Expression FixType(MSA.Expression expr)
    {
      if (expr.Type.IsValueType)
      {
        expr = Mast.DynamicConvert(expr, typeof(object));
      }
      return expr;
    }

    protected static void FillBody(CodeBlock cb, List<Statement> stmts, ExpressionList body)
    {
      FillBody(cb, stmts, body, true);
    }

    protected static void FillBody(CodeBlock cb, List<Statement> stmts, ExpressionList body, bool allowtailcall)
    {
      for (int i = 0; i < body.Count; i++)
      {
        Expression expr = body[i];
        MSA.Expression _expr = expr.GetAst(cb);
        if (i == body.Count - 1)
        {
          if (allowtailcall)
          {
            if (_expr is MethodCallExpression && _expr.Type != typeof(void))
            {
              ((MethodCallExpression)_expr).TailCall = true;
            }
            if (_expr is ActionExpression && _expr.Type != typeof(void))
            {
              ((ActionExpression)_expr).TailCall = true;
            }
          }
          stmts.Add(Mast.Return(_expr));
        }
        else
        {
          stmts.Add(Mast.Statement(expr.Span, _expr));
        }
      }

      cb.Body = Mast.Block(stmts.ToArray());
    }
  }

  #region CLR forms

  public class The : SpecialExpression
  {
    string type;
    Expression expr;

    public The(Expression type, Expression expr)
    {
      this.type = (type as Symbol).Name;
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      // TODO: make better converter
      Type t = TypeResolver.GetType(type);
      return Mast.DynamicConvert(expr.GetAst(cb), t);
    }
  }


  public class Try : SpecialExpression
  {
    Expression expr, @catch, @finally;

    public Try(Expression expr, Expression @catch, Expression @finally)
    {
      this.@catch = @catch;
      this.@finally = @finally;
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      Variable ct = cb.CreateTemporaryVariable(SymbolTable.StringToId("$exception"), typeof(Exception));
      if (@catch == null)
      {
        if (@finally == null)
        {
          return Mast.Void(Mast.TryFinally(Mast.Statement(expr.GetAst(cb)), null));
        }
        else
        {
          return Mast.Void(Mast.TryFinally(Mast.Statement(expr.GetAst(cb)), Mast.Statement(@finally.GetAst(cb))));
        }
      }
      else if (@finally == null)
      {
        CatchBlock[] cbs = new CatchBlock[] { Mast.Catch(typeof(Exception), ct, Mast.Statement(@catch.GetAst(cb))) };
        return Mast.Void(Mast.TryCatch(Span, SourceLocation.None, Mast.Statement(expr.GetAst(cb)), cbs));
      }
      else
      {
        CatchBlock[] cbs = new CatchBlock[] { Mast.Catch(typeof(Exception), ct, Mast.Statement(@catch.GetAst(cb))) };
        return Mast.Void(Mast.TryCatchFinally(Mast.Statement(expr.GetAst(cb)), cbs, Mast.Statement(@finally.GetAst(cb))));
      }
    }

    public override string ToString()
    {
      return "(try ...)";
    }
  }


  public class Call : SpecialExpression
  {
    string name;
    Expression target;
    List<Expression> args;

    public Call(Expression name, Expression target, IEnumerable<Expression> args)
    {
      this.name = (name as Symbol).Name;
      this.target = target;
      this.args = new List<Expression>(args);
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      List<Expression> args = new List<Expression>(this.args);
      args.Insert(0, this.target);
      BindingFlags si = 0;

      Symbol t = this.target as Symbol;

      Type target = t == null ? null : TypeResolver.GetType(t.Name);

      MSA.Expression instance = null;
      
      if (t == null || target == null)
      {
        si = BindingFlags.Instance;
        instance = this.target.GetAst(cb);
        target = instance.Type;
      }
      else
      {
        si = BindingFlags.Static;
      }
      if (target != null)
      {
        List<MethodBase> cans = new List<MethodBase>();
        foreach (MemberInfo mi in target.GetMember(name, MemberTypes.Method, si | BindingFlags.Public | 
          BindingFlags.FlattenHierarchy | BindingFlags.IgnoreCase))
        {
          MethodInfo methi = mi as MethodInfo;
          if (methi != null)
          {
            cans.Add(methi);
          }
        }

        if (cans.Count > 0)
        {
          MethodBinder mb =
            MethodBinder.MakeBinder(new IronLisp.Actions.IronLispActionBinder(null), name, cans.ToArray(), BinderType.Normal);

          CallType ct = CallType.ImplicitInstance;

          if (si == BindingFlags.Static)
          {
            ct = CallType.None;
            args.RemoveAt(0);
          }

          MSA.Expression[] aa = args.ConvertAll<MSA.Expression>(
            delegate(Expression e)
            {
              return e.GetAst(cb);
            }).ToArray();

          Type[] at = Array.ConvertAll<MSA.Expression, Type>(aa,
            delegate(MSA.Expression e) { return e.Type; });

          MethodCandidate mc = mb.MakeBindingTarget(ct, at);

          if (mc != null)
          {

            if (si == BindingFlags.Instance)
            {
              aa = ArrayUtils.RemoveFirst(aa);
            }

            return
              Mast.Call(instance, mc.Target.Method as MethodInfo, aa);
          }
        }
      }



      throw new NotImplementedException();
    }

    public override string ToString()
    {
      return "(call ...)";
    }
  }

  public class Is : SpecialExpression
  {
    string name;
    Expression expr;

    public Is(Expression name, Expression expr)
    {
      this.name = (name as Symbol).Name;
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      Type t = TypeResolver.GetType(name);
      Debug.Assert(t != null);
      return Mast.TypeIs(expr.GetAst(cb), t);
    }

    public override string ToString()
    {
      return string.Format("(is {0} {1})", name, expr);
    }
  }

  public class Throw : SpecialExpression
  {
    Expression expr;

    public Throw(Expression expr)
    {
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      if (expr == null)
      {
        return Mast.Rethrow();
      }
      else
      {
        return Mast.Throw(expr.GetAst(cb));
      }
    }

    public override string ToString()
    {
      return string.Format("(throw{0})", expr == null ? "" : (" " + expr));
    }
  }

  public class New : SpecialExpression
  {
    List<Expression> exprs;
    string name;

    public New(Expression name, IEnumerable<Expression> args)
    {
      this.exprs = new List<Expression>(args);
      this.name = (name as Symbol).Name;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      Type t = TypeResolver.GetType(name);

      MSA.Expression[] aa = exprs.ConvertAll<MSA.Expression>(
          delegate(Expression e) { return e.GetAst(cb); }).ToArray();

      Type[] args = new Type[exprs.Count];
      for (int i = 0; i < aa.Length; i++)
      {
        args[i] = aa[i].Type;
      }
      ConstructorInfo ci = t.GetConstructor(args);
      if (ci != null)
      {
        return Mast.New(ci, aa);
      }
      List<ConstructorInfo> cans = new List<ConstructorInfo>();
      foreach (ConstructorInfo pci in t.GetConstructors())
      {
        ParameterInfo[] pis = pci.GetParameters();
        if (pis.Length == args.Length)
        {
          cans.Add(pci);
        }
      }
      List<ConstructorInfo> cans2 = new List<ConstructorInfo>();
      foreach (ConstructorInfo pci in cans)
      {
        ParameterInfo[] pis = pci.GetParameters();
        bool paramsmatch = true;
        for (int i = 0; i < pis.Length; i++)
        {
          ParameterInfo pi = pis[i];
          Type pt = args[i];
          if (pi.ParameterType == pt || pt == typeof(object))
          {

          }
          else
          {
            paramsmatch = false;
          }
        }
        if (paramsmatch)
        {
          cans2.Add(pci);
        }
      }
      foreach (ConstructorInfo pci in cans2)
      {
        List<MSA.Expression> cargs = new List<MSA.Expression>();
        ParameterInfo[] pis = pci.GetParameters();

        for (int i = 0; i < args.Length; i++)
        {
          ParameterInfo pi = pis[i];
          Type pt = args[i];
          if (pt != pi.ParameterType)
          {
            cargs.Add(Mast.Convert(exprs[i].GetAst(cb), pi.ParameterType));
          }
          else
          {
            cargs.Add(exprs[i].GetAst(cb));
          }
        }
        return Mast.New(pci, cargs.ToArray());
      }
      return Mast.Action.Create(CreateInstanceAction.Make(aa.Length), t, aa);
    }

    public override string ToString()
    {
      if (exprs.Count == 0)
      {
        return string.Format("(new {0})", name);
      }
      else
      {
        return string.Format("(new {0} {1})", name, Join(exprs));
      }
    }
  }

  #endregion

  #region Derived forms

  public class Not : SpecialExpression
  {
    Expression expr;

    public Not(Expression expr)
    {
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      MSA.Expression final = expr.GetAst(cb);

      if (final.Type == typeof(bool))
      {
        return Mast.Not(final);
      }
      else
      {
        return Mast.Not(
          Mast.Call(null, typeof(IronLisp.Runtime.Builtins).GetMethod("IsTrue"), final)); ;
      }
    }

    public override string ToString()
    {
      return string.Format("(not {0})", expr);
    }
  }

  public class And : SpecialExpression
  {
    List<Expression> rest;

    public And(IEnumerable<Expression> rest)
    {
      this.rest = new List<Expression>(rest);
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      List<MSA.Expression> rhs = new List<MSA.Expression>();

      for (int i = 0; i < rest.Count; i++)
			{
        MSA.Expression lhs = null;
        if (i == 0)
        {
			    lhs = Mast.True();
        }
        else
        {
          lhs = rhs[i - 1];
        }
        rhs.Add(Mast.CoalesceTrue(rest[i].Span, cb, lhs, rest[i].GetAst(cb), typeof(Builtins).GetMethod("IsTrue")));
			}

      if (rhs.Count == 0)
      {
        rhs.Add( Mast.True());
      }
      

      MSA.Expression final = rhs[rhs.Count - 1];

      return final;
    }

    public override string ToString()
    {
      return string.Format("(and {0})", Join(rest));
    }
  }

  public class Or : SpecialExpression
  {
    ExpressionList rest;

    public Or(ExpressionList rest)
    {
      this.rest = rest;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      List<MSA.Expression> rhs = new List<MSA.Expression>();

      for (int i = 0; i < rest.Count; i++)
			{
        MSA.Expression lhs = null;
        if (i == 0)
        {
			    lhs = Mast.False();
        }
        else
        {
          lhs = rhs[i - 1];
        }
        rhs.Add(Mast.CoalesceFalse(rest[i].Span, cb, lhs, rest[i].GetAst(cb), typeof(Builtins).GetMethod("IsTrue")));
			}

      if (rhs.Count == 0)
      {
        rhs.Add(Mast.False());
      }

      MSA.Expression final = rhs[rhs.Count - 1];

      return final;
    }


    public override string ToString()
    {
      return string.Format("(or {0})", Join(rest));
    }
  }

  public class CondClause
  {
    readonly Expression test, result;

    public Expression Result
    {
      get { return result; }
    } 


    public Expression Test
    {
      get { return test; }
    } 


    public CondClause(Expression test, Expression result)
    {
      if (result == null)
      {
        this.result = test;
      }
      else if (result is ExpressionList)
      {
        ExpressionList r = result as ExpressionList;
        if (r.Count == 2 && r[0] is Symbol && ((Symbol)r[0]).Name == "=>")
        {
          this.result = new ExpressionList(r[1], test);
        }
        else
        {
          this.result = new Begin(result as ExpressionList);
        }
      }
      else
      {
        this.result = new ExpressionList(result, test);
      }
      this.test = test;
    }
  }


  /// <summary>
  /// Cond
  /// </summary>
  public class Cond : SpecialExpression
  {
    Expression @else;
    List<CondClause> tests;

    public Cond(IEnumerable<CondClause> tests, ExpressionList @else)
    {
      this.tests = new List<CondClause>(tests);
      if (@else != null)
      {
        this.@else = new Begin(@else);
      }
    }

    static MSA.Expression MakeCond(List<CondClause> tests, Expression @else, CodeBlock cb)
    {
      if (tests.Count == 0)
      {
        if (@else != null)
        {
          MSA.Expression r = @else.GetAst(cb);
          if (r.Type == typeof(void))
          {
            return Mast.Comma(r, Mast.Null());
          }
          return r;
        }
        else
        {
          return Mast.Null();
        }
      }
      else
      {
        CondClause test = tests[0];
        tests.RemoveAt(0);
        MSA.Expression testexpr = test.Test.GetAst(cb);
        if (testexpr.Type != typeof(bool))
        {
          testexpr = Mast.Call(null, typeof(Builtins).GetMethod("IsTrue"), testexpr);
        }
        return Mast.Condition(testexpr, test.Result.GetAst(cb), MakeCond(tests, @else, cb));
      }

    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      List<CondClause> copy = new List<CondClause>(tests);
      return MakeCond(copy, @else, cb);
    }

    public override string ToString()
    {
      return "(cond ...)";
    }
  }

  public class CaseClause
  {
    readonly Expression test, result;

    public Expression Result
    {
      get { return result; }
    } 


    public Expression Test
    {
      get { return test; }
    }


    public CaseClause(ExpressionList test, ExpressionList result)
    {
      this.result = new Begin(result);
      this.test = new Quote(test);
    }
  }

  public class Case : SpecialExpression
  {
    Expression key;
    Expression @else;
    List<CaseClause> tests;

    public Case(Expression key, List<CaseClause> tests, ExpressionList @else)
    {
      this.key = key;
      this.tests = tests;
      if (@else != null)
      {
        this.@else = new Begin(@else);
      }
    }

    static MSA.Expression MakeCase(Expression key, List<CaseClause> tests, Expression @else, CodeBlock cb)
    {
      if (tests.Count == 0)
      {
        if (@else != null)
        {
          return @else.GetAst(cb);
        }
        else
        {
          return Mast.Null();
        }
      }
      else
      {
        CaseClause test = tests[0];
        tests.RemoveAt(0);

        ExpressionList tl = new ExpressionList(new Symbol("memv"), key, test.Test);
        MSA.Expression testexpr = tl.GetAst(cb);

        if (testexpr.Type != typeof(bool))
        {
          testexpr = Mast.Call(null, typeof(Builtins).GetMethod("IsTrue"), testexpr);
        }

        return Mast.Condition(testexpr, test.Result.GetAst(cb), MakeCase(key, tests, @else, cb));
      }

    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      List<CaseClause> copy = new List<CaseClause>(tests);
      return MakeCase(key, copy, @else, cb);
    }

    public override string ToString()
    {
      return "(case ...)";
    }
  }


  // does this introduce a new lexical scope?
  public class Begin : SpecialExpression
  {
    List<Expression> exprs;

    public Begin(IEnumerable<Expression> exprs)
    {
      this.exprs = new List<Expression>(exprs);
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      if (exprs.Count == 0)
      {
        return Mast.Null();
      }
      int vi = exprs.Count - 1;
      return Mast.Comma(vi, exprs.ConvertAll<MSA.Expression>(delegate(Expression e)
      {
        return e.GetAst(cb);
      }));
    }

    public override string ToString()
    {
      return string.Format("(begin {0})", Join(exprs));
    }
  }

  public class DoClause
  {
    readonly Expression var, init, step;

    public Expression Step
    {
      get { return step; }
    } 


    public Expression Init
    {
      get { return init; }
    } 


    public Expression Var
    {
      get { return var; }
    } 


    public DoClause(Expression var, Expression init, Expression step)
    {
      this.var = var;
      this.init = init;
      this.step = step ?? var;
    }
  }

  public class Do : SpecialExpression
  {
    List<DoClause> doclauses;
    ExpressionList dotest, commands;

    public Do(List<DoClause> doclauses, ExpressionList dotest, ExpressionList commands)
    {
      this.doclauses = doclauses;
      this.dotest = dotest;
      this.commands = commands;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      CodeBlock wrap = Mast.CodeBlock(GetFullname(cb, "do"));
      wrap.Parent = cb;

      List<Statement> stmts = new List<Statement>();

      foreach (DoClause dc in doclauses)
      {
        MSA.Expression expr = dc.Init.GetAst(cb); // cheat?

        Variable v = Create(((Symbol)dc.Var).Name, wrap, expr.Type);
        Variable v2 = Create(((Symbol)dc.Var).Name + "::temp", wrap, expr.Type);
        stmts.Add(Mast.Write(v, FixType(expr)));
      }

      List<MSA.Expression> steps = new List<MSA.Expression>();

      foreach (DoClause dc in doclauses)
      {
        Variable v = FindVar(wrap, SymbolTable.StringToId(((Symbol)dc.Var).Name + "::temp"));
		    steps.Add(Mast.Assign(v, FixType(dc.Step.GetAst(wrap))));
      }

      foreach (DoClause dc in doclauses)
      {
        Variable v = FindVar(wrap, SymbolTable.StringToId(((Symbol)dc.Var).Name));
        Variable v2 = FindVar(wrap, SymbolTable.StringToId(((Symbol)dc.Var).Name + "::temp"));
        steps.Add(Mast.Void(Mast.Write(v, v2)));
      }

      MSA.Expression step = Mast.Comma(steps);

      stmts.Add(Mast.Loop(new Not(dotest[0]).GetAst(wrap), step, 
        Mast.Statement(new Begin(commands).GetAst(wrap)),
        Mast.Return(new Begin(dotest[1] as ExpressionList).GetAst(wrap))));

      stmts.Add(Mast.Return(Mast.Null()));

      FillBody(wrap, stmts, new ExpressionList());

      return Mast.Action.Call(typeof(object), MakeClosure(wrap));
    }
  }

  #endregion

  public class Macro : SpecialExpression
  {
    internal string namehint;
    ExpressionList args;
    ExpressionList body;

    internal ExpressionList Body
    {
      get { return body; }
    }

    internal ExpressionList Parameters
    {
      get { return args; }
    }

    public Macro(ExpressionList args, ExpressionList body)
    {
      this.args = args;
      this.body = body;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      string name = namehint;
      string cbname = GetFullname(c, name);

      CodeBlock cb = Mast.CodeBlock(cbname);

      bool isrest = false;


      foreach (Symbol s in args)
      {
        if (s.Name == ".")
        {
          isrest = true;
        }
        else
        {
          if (isrest)
          {
            CreateParameter(s.Name, cb, typeof(IEnumerable));
          }
          else
          {
            CreateParameter(s.Name, cb, typeof(object));
          }
        }
      }
      List<Statement> stmts = new List<Statement>();
      FillBody(cb, stmts, body);

      CodeBlockExpression cbe = Mast.CodeBlockExpression(cb, false);

      MSA.Expression ex = null;

      if (isrest)
      {
        ex = Mast.Call(null, typeof(Runtime.Macro).GetMethod("MakeVarArgX"), Mast.CodeContext(), cbe, Mast.Constant(cb.Parameters.Count), Mast.Constant(cb.Name));
      }
      else
      {
        ex = Mast.Call(null, typeof(Runtime.Macro).GetMethod("Make"), Mast.CodeContext(), cbe, Mast.Constant(cb.Name));
      }

      cb.BindClosures();

      CallTargetWithContextN md = cb.GetDelegateForInterpreter(Compiler, false) as CallTargetWithContextN;

      if (isrest)
      {
        Compiler.Scope.SetName(SymbolTable.StringToId(name), Runtime.Macro.MakeVarArgX(Compiler, md, cb.Parameters.Count, cb.Name));
      }
      else
      {
        Compiler.Scope.SetName(SymbolTable.StringToId(name), Runtime.Macro.Make(Compiler, md, cb.Name));
      }

      return ex;
    }

    public override string ToString()
    {
      return string.Format("(macro ({0}) {1})", Join(args), Join(body));
    }
  }

  public class Delay : SpecialExpression
  {
    Expression expr;

    public Delay(Expression expr)
    {
      this.expr = expr;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      CodeBlock c = Mast.CodeBlock(GetFullname(cb,"delay"));
      c.Parent = cb;

      FillBody(c, new List<Statement>(), new ExpressionList(expr));

      return Mast.Call(null,
        typeof(Promise).GetMethod("Make"), Mast.CodeContext(), Mast.CodeBlockExpression(c, false, false));

    }
  }

  #region Core forms

  public class Define : SpecialExpression
  {
    readonly Symbol name;

    public Symbol Name
    {
      get { return name; }
    }

    readonly Expression value;

    public Expression Value
    {
      get { return this.value; }
    } 



    public Define(Expression name, Expression value)
    {
      this.name = name as Symbol;
      this.value = value;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      Symbol s = name;
      Expression e = value;

      if (e is Lambda)
      {
        ((Lambda)e).namehint = name.Name;
      }

      if (e is Macro)
      {
        ((Macro)e).namehint = name.Name;
      }

      Variable v = Create(s.Name, cb, typeof(object));

      MSA.Expression me = e.GetAst(cb);

      return Mast.Void(Mast.Write(v, FixType(me)));
    }

    public override string ToString()
    {
      return string.Format("(define {0} {1})", name, value);
    }
  }

  public class Set : SpecialExpression
  {
    Symbol name;
    Expression value;
    
    public Set(Expression name, Expression value)
    {
      this.name = name as Symbol;
      this.value = value;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      Symbol s = name;
      Expression e = value;

      SymbolId sname = SymbolTable.StringToId(s.Name);
      Variable v = FindVar(cb, sname);

      MSA.Expression me = e.GetAst(cb);

      // not sure which way to go here
      if (v == null)
      {
        return Mast.Assign(sname, me);
      }
      else
      {
        return Mast.Void(Mast.Write(v, FixType(me)));
      }
    }

    public override string ToString()
    {
      return string.Format("(set! {0} {1})", name, value);
    }
  }

  public class If : SpecialExpression
  {
    Expression test, then, @else;

    public If(Expression test, Expression then, Expression @else)
    {
      this.test = test;
      this.then = then;
      this.@else = @else;
    }

    public override MSA.Expression GetAst(CodeBlock cb)
    {
      MSA.Expression e = null;
      if (@else != null)
      {
        e = @else.GetAst(cb);
      }
      else
      {
        e = Mast.Null();
      }

      return Mast.Condition(test.GetAst(cb), then.GetAst(cb), e, true);
    }

    public override string ToString()
    {
      if (@else == null)
      {
        return string.Format("(if {0} {1})", test, then);
      }
      else
      {
        return string.Format("(if {0} {1} {2})", test, then, @else);
      }
    }
  }

  public class Lambda : SpecialExpression
  {
    internal string namehint;
    ExpressionList args;
    ExpressionList body;
    internal CodeBlock cb;

    internal ExpressionList Body
    {
      get { return body; }
    }

    internal ExpressionList Parameters 
    {
      get { return args; } 
    }

    public Lambda(ExpressionList args, ExpressionList body)
    {
      this.args = args;
      this.body = body;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      string cbname = GetFullname(c, namehint ?? "lambda");

      CodeBlock cb = Mast.CodeBlock(cbname);
      cb.Parent = c;

      this.cb = cb;

      bool isrest = false;


      foreach (Symbol s in args)
      {
        if (s.Name == ".")
        {
          isrest = true;
        }
        else
        {
          if (isrest)
          {
            CreateParameter(s.Name, cb, typeof(IEnumerable));
          }
          else
          {
            CreateParameter(s.Name, cb, typeof(object));
          }
        }
      }

      List<Statement> stmts = new List<Statement>();
      FillBody(cb, stmts, body);

      MSA.Expression ex = null;

      if (isrest)
      {
        ex =
          Mast.Call(null, typeof(Closure).GetMethod("MakeVarArgX"), Mast.CodeContext(),
          Mast.CodeBlockExpression(cb, false, false), Mast.Constant(cb.Parameters.Count), Mast.Constant(cb.Name));
      }
      else
      {
        ex = MakeClosure(cb);
      }

      if (namehint != null)
      {
        //cb.Parent = null;
        cb.BindClosures();
        if (isrest)
        {
          Compiler.Scope.SetName(SymbolTable.StringToId(namehint), Closure.MakeVarArgX(Compiler, cb.GetDelegateForInterpreter(Compiler, false), cb.Parameters.Count, cb.Name));
        }
        else
        {
          Compiler.Scope.SetName(SymbolTable.StringToId(namehint), Closure.Make(Compiler, cb.GetDelegateForInterpreter(Compiler, false), cb.Name));
        }
      }

      return ex;
    }



    public override string ToString()
    {
      if (args.Count == 2 && ((Symbol) args[0]).Name == ".")
      {
        return string.Format("(lambda {0} {1})", args[1], Join(body));
      }
      return string.Format("(lambda ({0}) {1})", Join(args), Join(body));
    }
  }

  #endregion

  #region Let and friends

  public class Let : SpecialExpression
  {
    ExpressionList assigns, stmts;

    public Let(ExpressionList assigns, ExpressionList stmts)
    {
      this.assigns = assigns;
      this.stmts = stmts;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      CodeBlock cb = Mast.CodeBlock(GetFullname(c, "let"));
      cb.Parent = c;

      List<Statement> stmts = new List<Statement>();

      List<MSA.Expression> args = new List<MSA.Expression>();
      
      for (int i = 0; i < assigns.Count; i++)
      {
        ExpressionList ass = assigns[i] as ExpressionList;
        Symbol s = ass[0] as Symbol;
        Expression value = ass[1];

        if (value is Lambda)
        {
          ((Lambda)value).namehint = s.Name;
        }


        MSA.Expression ex = value.GetAst(c);
        args.Add(ex);
        Variable v = CreateParameter(s.Name, cb, ex.Type);
      }

      ExpressionList body = this.stmts;

      FillBody(cb, stmts, body);

      args.Insert(0, MakeClosure(cb));

      return Mast.Action.Call(typeof(object), args.ToArray());
    }

 

    public override string ToString()
    {
      return string.Format("(let ({0}) {1})", Join(assigns), Join(stmts));
    }
  }

  public class LetStar : SpecialExpression
  {
    ExpressionList assigns, stmts;

    public LetStar(ExpressionList assigns, ExpressionList stmts)
    {
      this.assigns = assigns;
      this.stmts = stmts;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      CodeBlock cb = Mast.CodeBlock(GetFullname(c, "let*"));
      cb.Parent = c;

      List<Statement> stmts = new List<Statement>();

      for (int i = 0; i < assigns.Count; i++)
      {
        ExpressionList ass = assigns[i] as ExpressionList;
        Symbol s = ass[0] as Symbol;
        Expression value = ass[1];

        if (value is Lambda)
        {
          ((Lambda)value).namehint = s.Name;
        }

        MSA.Expression ex = value.GetAst(cb);
        Variable v = Create(s.Name, cb, ex.Type);

        stmts.Add(Mast.Statement(Mast.Assign(v, FixType(ex))));
      }

      FillBody(cb, stmts, this.stmts);

      return Mast.Action.Call(typeof(object), MakeClosure(cb));
    }

    public override string ToString()
    {
      return string.Format("(let* ({0}) {1})", Join(assigns), Join(stmts));
    }
  }

  public class LetRec : SpecialExpression
  {
    ExpressionList assigns, stmts;

    public LetRec(ExpressionList assigns, ExpressionList stmts)
    {
      this.assigns = assigns;
      this.stmts = stmts;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      CodeBlock wrapper = Mast.CodeBlock(GetFullname(c, "letrec::wrapper"));
      wrapper.Parent = c;

      CodeBlock cb = Mast.CodeBlock(GetFullname(c, "letrec"));
      cb.Parent = wrapper;

      List<Statement> stmts = new List<Statement>();
      List<Statement> wrapstmts = new List<Statement>();
      List<Variable> vars = new List<Variable>();

      for (int i = 0; i < assigns.Count; i++)
      {
        ExpressionList ass = assigns[i] as ExpressionList;
        Symbol s = ass[0] as Symbol;
        Expression value = ass[1];

        if (value is Lambda)
        {
          ((Lambda)value).namehint = s.Name;
        }

        Variable v = Create(s.Name, wrapper, typeof(object));
        wrapstmts.Add(Mast.Write(v, Mast.ReadField(null, typeof(Uninitialized), "Instance")));

        vars.Add(v);
      }

      for (int i = 0; i < assigns.Count; i++)
      {
        ExpressionList ass = assigns[i] as ExpressionList;
        Symbol s = ass[0] as Symbol;
        Expression value = ass[1];
        
        MSA.Expression ex = value.GetAst(wrapper);
        Variable vw = Create(s.Name + "::renamed", wrapper, ex.Type);
        wrapstmts.Add(Mast.Write(vw, FixType(ex)));

        stmts.Add(Mast.Write(vars[i], vw));
      }

      FillBody(cb, stmts, this.stmts);

      ActionExpression ae = Mast.Action.Call(typeof(object), MakeClosure(cb));

      if (ae.Type != typeof(void))
      {
        ae.TailCall = true;
      }

      wrapstmts.Add(Mast.Return(ae));

      wrapper.Body = Mast.Block(wrapstmts);

      return
        Mast.Action.Call(typeof(object), MakeClosure(wrapper));
    }

    public override string ToString()
    {
      return string.Format("(letrec ({0}) {1})", Join(assigns), Join(stmts));
    }
  }

  public class NamedLet : SpecialExpression
  {
    Symbol name;
    ExpressionList assigns, stmts;

    public NamedLet(Expression name, ExpressionList assigns, ExpressionList stmts)
    {
      this.name = name as Symbol;
      this.assigns = assigns;
      this.stmts = stmts;
    }

    public override MSA.Expression GetAst(CodeBlock c)
    {
      CodeBlock wrapper = Mast.CodeBlock(GetFullname(c, name.Name + "::wrapper"));
      wrapper.Parent = c;

      CodeBlock cb = Mast.CodeBlock(GetFullname(c, name.Name));
      cb.Parent = wrapper;

      Variable namedlet = Create(name.Name, cb, typeof(FastCallable));
      Variable renamedlet = Create(name.Name + "::renamed", wrapper, typeof(FastCallable));

      List<Statement> stmts = new List<Statement>();

      stmts.Add(Mast.Write(namedlet, renamedlet));

      List<MSA.Expression> args = new List<MSA.Expression>();

      for (int i = 0; i < assigns.Count; i++)
      {
        ExpressionList ass = assigns[i] as ExpressionList;
        Symbol s = ass[0] as Symbol;
        Expression value = ass[1];

        if (value is Lambda)
        {
          ((Lambda)value).namehint = s.Name;
        }

        MSA.Expression ex = value.GetAst(c);
        args.Add(ex);

        Variable v = CreateParameter(s.Name, cb, ex.Type);
      }

      FillBody(cb, stmts, this.stmts);

      args.Insert(0, Mast.Assign(renamedlet, MakeClosure(cb)));

      ActionExpression ae = Mast.Action.Call(typeof(object), args.ToArray());

      if (ae.Type != typeof(void))
      {
        ae.TailCall = true;
      }

      wrapper.Body = Mast.Return(ae);

      return
        Mast.Action.Call( typeof(object), MakeClosure(wrapper));
    }

    public override string ToString()
    {
      return string.Format("(let {2} ({0}) {1})", Join(assigns), Join(stmts), name);
    }
  }

  #endregion

}
