#region License
/* ****************************************************************************
 * Copyright (c) Llewellyn Pritchard. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. 
 * A copy of the license can be found in the License.html file at the root of this distribution. 
 * By using this source code in any fashion, you are agreeing to be bound by the terms of the 
 * Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 * ***************************************************************************/
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using gppg;
using IronLisp.Ast;
using System.Collections;

namespace IronLisp.Compilers.Stage2
{
  public class Scanner : IScanner<ValueType, LexLocation>
  {
    LexLocation __yylloc;

    public override LexLocation yylloc 
    { 
      get { return __yylloc; } 
      set { __yylloc = value; } 
    }

    public override void yyerror(string format, params object[] args)
    {
      Console.Error.WriteLine(format, args);
    }

    readonly static Dictionary<string, Tokens> SymbolMap = new Dictionary<string, Tokens>(StringComparer.InvariantCultureIgnoreCase);

    static Scanner()
    {
      SymbolMap["and"] = Tokens.AND;
      SymbolMap["set!"] = Tokens.SET;
      SymbolMap["quasiquote"] = Tokens.QUASIQUOTE;
      SymbolMap["call"] = Tokens.CALL;
      SymbolMap["case"] = Tokens.CASE;
      SymbolMap["cond"] = Tokens.COND;
      SymbolMap["=>"] = Tokens.ARROW;
      SymbolMap["define"] = Tokens.DEFINE;
      SymbolMap["delay"] = Tokens.DELAY;
      SymbolMap["begin"] = Tokens.BEGIN;
      SymbolMap["do"] = Tokens.DO;
      SymbolMap["else"] = Tokens.ELSE;
      SymbolMap["lambda"] = Tokens.LAMBDA;
      SymbolMap["if"] = Tokens.IF;
      SymbolMap["is"] = Tokens.IS;
      SymbolMap["let"] = Tokens.LET;
      SymbolMap["let*"] = Tokens.LETSTAR;
      SymbolMap["letrec"] = Tokens.LETREC;
      SymbolMap["define-macro"] = Tokens.DEFINEMACRO;
      SymbolMap["macro"] = Tokens.MACRO;
      SymbolMap["new"] = Tokens.NEW;
      SymbolMap["not"] = Tokens.NOT;
      SymbolMap["or"] = Tokens.OR;
      SymbolMap["quote"] = Tokens.QUOTE;
      SymbolMap["unquote-splicing"] = Tokens.UNQUOTESPLICING;
      SymbolMap["reference"] = Tokens.REFERENCE;
      SymbolMap["the"] = Tokens.THE;
      SymbolMap["throw"] = Tokens.THROW;
      SymbolMap["try"] = Tokens.TRY;
      SymbolMap["unquote"] = Tokens.UNQUOTE;
      SymbolMap["using"] = Tokens.USING;
    }

    IEnumerable source;
    IEnumerator iter;

    public void SetSource(IEnumerable source)
    {
      this.source = source;
      iter = new DeepEnumerable(source).GetEnumerator();
    }

    public bool readascontent = false;


    int yywrap()
    {
      if (iter.MoveNext())
      {
        object cur = iter.Current;
        if (cur is Tokens)
        {
          return (int)((Tokens)cur);
        }
        else if (cur is Symbol)
        {
          Symbol s = (Symbol)cur;
          yylval.elem = s;
          yylloc = s.Location;
          Tokens tok;
          if (!readascontent && last == (int) Tokens.LBRACE && SymbolMap.TryGetValue(s.Name, out tok))
          {
            return (int)tok;
          }
          else
          {
            if (s.Name == ".")
            {
              return (int)Tokens.DOT;
            }
            else
            {
              return (int)Tokens.IDENTIFIER;
            }
          }
        }
        else if (cur is Literal)
        {
          Literal lit = (Literal)cur;
          yylval.elem = lit;
          yylloc = lit.Location;

          return (int)Tokens.LITERAL;
        }
        else if (cur is Expression)
        {
          Expression exp = (Expression)cur;
          yylval.elem = exp;
          yylloc = exp.Location;
          return (int)Tokens.EXPR;
        }
      }
      return (int)Tokens.EOF;

    }

    int last;

    public override int yylex()
    {
      int i = yywrap();
      Tokens t = (Tokens)i;
      return (last = i);
    }

    class DeepEnumerable : IEnumerable
    {
      IEnumerable source;

      public DeepEnumerable(IEnumerable source)
      {
        this.source = source;
      }

      static DeepEnumerable()
      {
        specops.Add("backquote", "backquote");
        specops.Add("quote", "quote");
        specops.Add("unquote", "unquote");
        specops.Add("splice", "splice");
      }

      static Dictionary<string, string> specops = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);

      public IEnumerator GetEnumerator()
      {
        int counter = 0;
        foreach (object obj in source)
        {
          counter++;
          if (source is ImproperList && ((ImproperList)source).Count == counter)
          {
            yield return Tokens.DOT;
          }
          ExpressionList exprlist = obj as ExpressionList;
          bool wasspec = false;
          if (exprlist != null)
          {
            if (exprlist.Count == 2 && exprlist[0] is Symbol)
            {
              Symbol s = exprlist[0] as Symbol;

              if (specops.ContainsKey(s.Name))
              {
                wasspec = true;
                yield return Tokens.LBRACE;
                yield return s;
                yield return exprlist[1];
                yield return Tokens.RBRACE;
              }
            }
          }

          if (!wasspec)
          {
            IEnumerable inner = obj as IEnumerable;
            if (inner != null)
            {
              yield return Tokens.LBRACE;
              foreach (object innerobj in new DeepEnumerable(inner))
              {
                yield return innerobj;
              }
              yield return Tokens.RBRACE;
            }

            else
            {
              yield return obj;
            }
          }
        }
      }
    }
  }
}
